# compute derivative for bi-level optimization
import numpy as np 
import jax.numpy as jnp
from jax import grad, jit, vmap, jacfwd, jacrev, random, lax
from functools import partial
import jax
from trilevel.implicit_bilevel import find_hx
from optimizer import sgd, adam

def find_gx(f, x, y, z, step_size=0.01, max_iters=10):
    grad_f = grad(f, argnums=2)
    for _ in range(max_iters):
        z -= step_size * grad_f(x, y, z)
    return z

# Function to compute dg(x)/dx
def dg(f, x, y, z):
    dgdx = -jacrev(jacrev(f, argnums=2), argnums=0)(x, y, z) / jacrev(jacrev(f, argnums=2), argnums=2)(x, y, z)
    dgdy = -jacrev(jacrev(f, argnums=2), argnums=1)(x, y, z) / jacrev(jacrev(f, argnums=2), argnums=2)(x, y, z)
    numerator_dxdy = dgdx * dgdy * jacrev(jacrev(jacrev(f, argnums=2), argnums=2), argnums=2)(x, y, z) +\
        dgdy * jacrev(jacrev(jacrev(f, argnums=2), argnums=0), argnums=2)(x, y, z) +\
        jacrev(jacrev(jacrev(f, argnums=2), argnums=0), argnums=1)(x, y, z) +\
        dgdx *jacrev(jacrev(jacrev(f, argnums=2), argnums=1), argnums=2)(x, y, z)
    d2gdxdy = -numerator_dxdy / jacrev(jacrev(f, argnums=2), argnums=2)(x, y, z)
    numerator_dy2 = dgdy ** 2 * jacrev(jacrev(jacrev(f, argnums=2), argnums=2), argnums=2)(x, y, z) +\
        2 * dgdy * jacrev(jacrev(jacrev(f, argnums=2), argnums=1), argnums=2)(x, y, z) +\
        jacrev(jacrev(jacrev(f, argnums=2), argnums=1), argnums=1)(x, y, z)
    d2gdy2 = -numerator_dy2 / jacrev(jacrev(f, argnums=2), argnums=2)(x, y, z)
    return dgdy, dgdx, d2gdxdy, d2gdy2

def dh_dx(f, x, y, z, dgdy, dgdx, d2gdxdy, d2gdy2):
    numerator = jacrev(jacrev(f, argnums=0), argnums=1)(x, y, z) + dgdx * jacrev(jacrev(f, argnums=1), argnums=2)(x, y, z) +\
       dgdy * jacrev(jacrev(f, argnums=0), argnums=2)(x, y, z) + dgdx * dgdy * jacrev(jacrev(f, argnums=2), argnums=2)(x, y, z) +\
       d2gdxdy * jacrev(f, argnums=2)(x, y, z)
    denominator = jacrev(jacrev(f, argnums=1), argnums=1)(x, y, z) + 2 * dgdy * jacrev(jacrev(f, argnums=1), argnums=2)(x, y, z) +\
       dgdy ** 2 * jacrev(jacrev(f, argnums=2), argnums=2)(x, y, z) + d2gdy2 * jacrev(f, argnums=2)(x, y, z)
    return -numerator / denominator

# total derivative of f1(x, y, z) w.r.t. x
def df_dx(f1, f2, f3, x, y, z):
    dgdy, dgdx, d2gdxdy, d2gdy2 = dg(f3, x, y, z)
    dhdx = dh_dx(f2, x, y, z, dgdy, dgdx, d2gdxdy, d2gdy2)
    return jacrev(f1, argnums=0)(x, y, z) + jacrev(f1, argnums=1)(x, y, z) * dhdx +\
          jacrev(f1, argnums=2)(x, y, z) * (dgdx + dgdy * dhdx)

def wrap_df(f1, f2, f3, x, y, z):
    y, z = find_hx(f2, f3, jnp.stack([x,y,z], axis=0), max_iters=10)
    z = find_gx(f3, x, y, z, max_iters=10)
    dx = df_dx(f1, f2, f3, x, y, z)
    return dx, y, z